{
MATRICES.PAS

Unit for matrix and vector manipulation.
A matrix is treated as a set of vectors.

Don't even try to understand the pointer operations.

(C) C. Valens, <c.valens@mindless.com>

Created     : 10/04/1998
Last update : 29/10/1998
}


unit matrices;

{******************************************************}
{*                     INTERFACE                      *}
{******************************************************}
interface

type
{ Define the data type to be stored in the matrix and vectors.}
  element_type = integer;

{ Vector type definitions. }
  element_ptr = ^element_type;
  vector = record
             v: element_ptr;
             l: integer;
           end;

{ Matrix type definitions. }
  vector_ptr = ^vector;
  matrix = record
             m: vector_ptr;
             r: integer;
             c: integer;
           end;

  statistics = record
                 N: real;
                 mean: real;
                 variance: real;
               end;

const
  min_element_type = -32768;
  max_element_type =  32767;

{ Vector operations. }
procedure create_vector(var v: vector; l: integer; var error: boolean);
procedure destroy_vector(var v: vector);
function get_vector_element(v: vector; i: integer): element_type;
procedure put_vector_element(var v: vector; i: integer; d: element_type);
function vector_max(v: vector): element_type;
function abs_vector_max(v: vector): element_type;
function vector_min(v: vector): element_type;
procedure copy_vector(src: vector; var dest: vector);
procedure write_vector(v: vector);
procedure clear_vector(var v: vector);
procedure absolute_clip_vector(var v: vector; clip_level: element_type);
procedure clip_vector(var v: vector; clip_level: element_type);
procedure absolute_threshold_vector(var v: vector; threshold: element_type);
procedure threshold_vector(var v: vector; threshold: element_type);
function vector_mean(v: vector): real;
procedure vector_statistics(v: vector; var stats: statistics);

{ Matrix operations. }
procedure create_matrix(var m: matrix; r, c: integer; var error: boolean);
procedure destroy_matrix(var m: matrix);
function get_matrix_element(m: matrix; r, c: integer): element_type;
procedure put_matrix_element(var m: matrix; r, c: integer; d: element_type);
procedure get_row(m: matrix; r: integer; var v: vector);
procedure put_row(var m: matrix; r: integer; v: vector);
procedure get_column(m: matrix; c: integer; var v: vector);
procedure put_column(var m: matrix; c: integer; v: vector);
function matrix_max(m: matrix): element_type;
function abs_matrix_max(m: matrix): element_type;
function matrix_min(m: matrix): element_type;
procedure copy_matrix(src: matrix; var dest: matrix);
procedure write_matrix(m: matrix);
procedure normalize_matrix(var m: matrix; norm: real);
procedure unnormalize_matrix(var m: matrix; min, max: element_type);
procedure reorder_matrix(src: matrix; var dest: matrix; max_level: integer);
procedure deorder_matrix(src: matrix; var dest: matrix; max_level: integer);
procedure test_matrix(m: matrix);
procedure clear_matrix(var m: matrix);
procedure absolute_clip_matrix(var m: matrix; clip_level: element_type);
procedure clip_matrix(var m: matrix; clip_level: element_type);
procedure absolute_threshold_matrix(var m: matrix; threshold: element_type);
procedure threshold_matrix(var m: matrix; threshold: element_type);
function matrix_psnr(m, ref: matrix): real;
function matrix_mean(m: matrix): real;
procedure matrix_statistics(m: matrix; var stats: statistics);
procedure matrix_shift(m: matrix; shift: element_type);


{******************************************************}
{*                  IMPLEMENTATION                    *}
{******************************************************}
implementation

{******************************************************}
{*                Vector manipulations                *}
{******************************************************}
procedure create_vector(var v: vector; l: integer; var error: boolean);
begin
  error := FALSE;
(*  GetMem(v.v,l*SizeOf(element_ptr));*)
  GetMem(v.v,l*SizeOf(element_type));
  if v.v=NIL then begin
    error := TRUE;
    v.l := 0;
  end
  else v.l := l;
end;

procedure destroy_vector(var v: vector);
begin
(*  if v.v<>NIL then FreeMem(v.v,v.l*SizeOf(element_ptr));*)
  if v.v<>NIL then FreeMem(v.v,v.l*SizeOf(element_type));
  v.l := 0;
end;

function get_vector_element(v: vector; i: integer): element_type;
begin
  get_vector_element := element_ptr(longint(v.v)+i*SizeOf(element_type))^;
end;

procedure put_vector_element(var v: vector; i: integer; d: element_type);
begin
  element_ptr(longint(v.v)+i*SizeOf(element_type))^ := d;
end;

function vector_max(v: vector): element_type;
var
  i: integer;
  max, temp: element_type;
begin
  max := get_vector_element(v,0);
  for i:=0 to v.l-1 do begin
    temp := get_vector_element(v,i);
    if max<temp then max := temp;
  end;
  vector_max := max;
end;

function abs_vector_max(v: vector): element_type;
var
  i: integer;
  max, temp: element_type;
begin
  max := 0;
  for i:=0 to v.l-1 do begin
    temp := Abs(get_vector_element(v,i));
    if max<temp then max := temp;
  end;
  abs_vector_max := max;
end;

function vector_min(v: vector): element_type;
var
  i: integer;
  min, temp: element_type;
begin
  min := get_vector_element(v,0);
  for i:=0 to v.l-1 do begin
    temp := get_vector_element(v,i);
    if min>temp then min := temp;
  end;
  vector_min := min;
end;

procedure copy_vector(src: vector; var dest: vector);
var
  i: integer;
begin
  for i:=0 to src.l-1 do begin
    put_vector_element(dest,i,get_vector_element(src,i));
  end;
end;

procedure write_vector(v: vector);
var
  i: integer;
begin
  for i:=0 to v.l-1 do begin
    write(get_vector_element(v,i):3);
  end;
end;

procedure clear_vector(var v: vector);
var
  i: integer;
begin
  for i:=0 to v.l-1 do begin
    put_vector_element(v,i,0);
  end;
end;

procedure absolute_clip_vector(var v: vector; clip_level: element_type);
var
  i: integer;
begin
  for i:=0 to v.l-1 do begin
    if Abs(get_vector_element(v,i))>clip_level then begin
      put_vector_element(v,i,clip_level);
    end
  end;
end;

procedure clip_vector(var v: vector; clip_level: element_type);
var
  i: integer;
begin
  for i:=0 to v.l-1 do begin
    if get_vector_element(v,i)>clip_level then begin
      put_vector_element(v,i,clip_level);
    end
  end;
end;

procedure absolute_threshold_vector(var v: vector; threshold: element_type);
var
  i: integer;
begin
  for i:=0 to v.l-1 do begin
    if Abs(get_vector_element(v,i))<=threshold then begin
      put_vector_element(v,i,0);
    end
  end;
end;

procedure threshold_vector(var v: vector; threshold: element_type);
var
  i: integer;
begin
  for i:=0 to v.l-1 do begin
    if get_vector_element(v,i)<=threshold then begin
      put_vector_element(v,i,0);
    end
  end;
end;

function vector_mean(v: vector): real;
var
  i: integer;
  mean, temp: real;
begin
  mean := 0;
  for i:=0 to v.l-1 do begin
    mean := mean + get_vector_element(v,i);
  end;
  temp := i;
  if temp<>0 then vector_mean := mean/temp
  else vector_mean := -1;
end;

procedure vector_statistics(v: vector; var stats: statistics);
var
  i: integer;
  data, mean, mean2: real;
begin
  mean := 0;
  mean2 := 0;
  stats.N := v.l;
  if stats.N<>0 then begin
    for i:=0 to v.l-1 do begin
      data := get_vector_element(v,i);
      mean := mean + data;
      mean2 := mean2 + Sqr(data);
    end;
    mean := mean/stats.N;
    mean2 := mean2/stats.N;
    stats.mean := mean;
    stats.variance := mean2 - Sqr(mean);
  end
  else begin
    stats.mean := 0;
    stats.variance := -1;
  end;
end;


{******************************************************}
{*                Matrix manipulations                *}
{******************************************************}
procedure create_matrix(var m: matrix; r, c: integer; var error: boolean);
var
  i: integer;
  temp: element_ptr;
begin
  error := FALSE;
  GetMem(m.m,r*SizeOf(vector));
  if m.m=NIL then error := TRUE
  else begin
    for i:=0 to r-1 do begin
      create_vector(vector_ptr(longint(m.m)+i*sizeof(vector))^,c,error);
    end;
  end;
  if error=FALSE then begin
    m.r := r;
    m.c := c;
  end
  else begin
    m.r := 0;
    m.c := 0;
  end;
end;

procedure destroy_matrix(var m: matrix);
var
  i: integer;
  temp: element_ptr;
begin
  for i:=0 to m.r-1 do begin
    destroy_vector(vector_ptr(longint(m.m)+i*sizeof(vector))^);
  end;
  if m.m<>NIL then FreeMem(m.m,m.r*SizeOf(vector));
  m.r := 0;
  m.c := 0;
end;

function get_matrix_element(m: matrix; r, c: integer): element_type;
begin
(*  get_matrix_element := get_vector_element(vector_ptr(longint(m.m)+r*sizeof(vector))^,c);*)
  get_matrix_element := element_ptr(longint(vector_ptr(longint(m.m)+r*sizeof(vector))^.v)+c*SizeOf(element_type))^;
end;

procedure put_matrix_element(var m: matrix; r, c: integer; d: element_type);
begin
(*  put_vector_element(vector_ptr(longint(m.m)+r*sizeof(vector))^,c,d);*)
  element_ptr(longint(vector_ptr(longint(m.m)+r*sizeof(vector))^.v)+c*SizeOf(element_type))^ := d;
end;

procedure get_row(m: matrix; r: integer; var v: vector);
var
  row: vector;
  i, N: integer;
begin
  row := vector_ptr(longint(m.m)+r*sizeof(vector))^;
  N := v.l;
  if v.l > m.c then N := m.c;
  for i:=0 to N-1 do begin
(*    put_vector_element(v,i,get_vector_element(row,i));*)
    put_vector_element(v,i,element_ptr(longint(row.v)+i*SizeOf(element_type))^);
  end;
end;

procedure put_row(var m: matrix; r: integer; v: vector);
var
  row: vector;
  i, N: integer;
begin
  row := vector_ptr(longint(m.m)+r*sizeof(vector))^;
  N := v.l;
  if v.l > m.c then N := m.c;
  for i:=0 to N-1 do begin
(*    put_vector_element(row,i,get_vector_element(v,i));*)
    put_vector_element(row,i,element_ptr(longint(v.v)+i*SizeOf(element_type))^);
  end;
end;

procedure get_column(m: matrix; c: integer; var v: vector);
var
  i, N: integer;
begin
  N := v.l;
  if v.l > m.r then N := m.r;
  for i:=0 to N-1 do begin
    put_vector_element(v,i,get_matrix_element(m,i,c));
  end;
end;

procedure put_column(var m: matrix; c: integer; v: vector);
var
  i, N: integer;
begin
  N := v.l;
  if v.l > m.r then N := m.r;
  for i:=0 to N-1 do begin
    put_matrix_element(m,i,c,get_vector_element(v,i));
  end;
end;

function matrix_max(m: matrix): element_type;
var
  row: vector;
  i: integer;
  max, temp: element_type;
  error: boolean;
begin
  create_vector(row,m.c,error);
  max := get_matrix_element(m,0,0);
  for i:=0 to m.r-1 do begin
    get_row(m,i,row);
    temp := vector_max(row);
    if max<temp then max := temp;
  end;
  destroy_vector(row);
  matrix_max := max;
end;

function abs_matrix_max(m: matrix): element_type;
var
  row: vector;
  i: integer;
  max, temp: element_type;
  error: boolean;
begin
  create_vector(row,m.c,error);
  max := 0;
  for i:=0 to m.r-1 do begin
    get_row(m,i,row);
    temp := abs_vector_max(row);
    if max<temp then max := temp;
  end;
  destroy_vector(row);
  abs_matrix_max := max;
end;

function matrix_min(m: matrix): element_type;
var
  row: vector;
  i: integer;
  min, temp: element_type;
  error: boolean;
begin
  create_vector(row,m.c,error);
  min := get_matrix_element(m,0,0);
  for i:=0 to m.r-1 do begin
    get_row(m,i,row);
    temp := vector_min(row);
    if min>temp then min := temp;
  end;
  destroy_vector(row);
  matrix_min := min;
end;

procedure copy_matrix(src: matrix; var dest: matrix);
var
  i: integer;
begin
  if (src.r=dest.r) and (src.c=dest.c) then begin
    for i:=0 to src.r-1 do begin
      get_row(src,i,vector_ptr(longint(dest.m)+i*SizeOf(vector))^);
    end;
  end;
end;

procedure write_matrix(m: matrix);
var
  i: integer;
begin
  for i:=0 to m.r-1 do begin
    write_vector(vector_ptr(longint(m.m)+i*SizeOf(vector))^);
    writeln;
  end;
end;

procedure normalize_matrix(var m: matrix; norm: real);
var
  min, max: element_type;
  i, j: integer;
  temp: real;
begin
  min := matrix_min(m);
  max := matrix_max(m);
  for i:=0 to m.r-1 do begin
    for j:=0 to m.c-1 do begin
      temp := get_matrix_element(m,i,j);
      if (max-min)>0 then begin
        temp := (temp-min)/(max-min)*norm;
      end
      else begin
        temp := norm;
      end;
      put_matrix_element(m,i,j,Round(temp));
    end;
  end;
end;

procedure unnormalize_matrix(var m: matrix; min, max: element_type);
var
  i, j: integer;
  temp: real;
  norm: element_type;
begin
  norm := matrix_max(m);
  for i:=0 to m.r-1 do begin
    for j:=0 to m.c-1 do begin
      temp := get_matrix_element(m,i,j);
      temp := (temp/norm)*(max-min)+min;
      put_matrix_element(m,i,j,Round(temp));
    end;
  end;
end;

procedure reorder_matrix(src: matrix; var dest: matrix; max_level: integer);
var
  x, x_max, y, y_offset, rd, cd, cd_max: integer;
  step, d: integer;
  level, hlevel, vlevel: integer;
  data: element_type;
begin
  step := 2;
  d := step shr 1;
  rd := src.r-1;
  cd_max := src.c-1;
  hlevel := src.c shr 1;
  vlevel := src.r shr 1;
  level := max_level;
  y_offset := 0;

  put_matrix_element(dest,0,0,get_matrix_element(src,0,0));

  while (level<>0) and (hlevel<>0) and (vlevel<>0) do begin
    for y:=(vlevel shl 1)-1 downto vlevel do begin
      cd := cd_max;
      for x:=(hlevel shl 1)-1 downto hlevel do begin

        data := get_matrix_element(src,rd,cd);(* diagonal coefficients. *)
        put_matrix_element(dest,y,x,data);

        data := get_matrix_element(src,rd,cd-d);(* horizontal coefficients. *)
        put_matrix_element(dest,y,x-hlevel,data);

        data := get_matrix_element(src,rd-d,cd);(* vertical coefficients. *)
        put_matrix_element(dest,y-vlevel,x,data);

        dec(cd,step);
      end;
      dec(rd,step);
    end;
    rd := src.r-1-(step-1);
    cd_max := src.c-(step-1)-1;
    d := step;
    step := step shl 1;
    inc(y_offset,vlevel);
    hlevel := hlevel shr 1;
    vlevel := vlevel shr 1;
    dec(level);
  end;

  if (hlevel<>0) and (vlevel<>0) then begin
    hlevel := src.c shr max_level;
    vlevel := src.r shr max_level;
    y_offset := src.r - vlevel;
    rd := 0;
    for y:=0 to vlevel-1 do begin
      cd := 0;
      for x:=0 to hlevel-1 do begin

        data := get_matrix_element(src,rd,cd);
        put_matrix_element(dest,y,x,data);

        inc(cd,d);
      end;
      inc(rd,d);
    end;
  end;

end;

procedure deorder_matrix(src: matrix; var dest: matrix; max_level: integer);
var
  x, x_max, y, y_offset, rd, cd, cd_max: integer;
  step, d: integer;
  level, hlevel, vlevel: integer;
  data: element_type;
begin
  step := 2;
  d := step shr 1;
  rd := src.r-1;
  cd_max := src.c-1;
  hlevel := src.c shr 1;
  vlevel := src.r shr 1;
  level := max_level;
  y_offset := 0;

  put_matrix_element(dest,0,0,get_matrix_element(src,0,0));

  while (level<>0) and (hlevel<>0) and (vlevel<>0) do begin
    for y:=(vlevel shl 1)-1 downto vlevel do begin
      cd := cd_max;
      for x:=(hlevel shl 1)-1 downto hlevel do begin

        data := get_matrix_element(src,y,x);(* diagonal coefficients. *)
        put_matrix_element(dest,rd,cd,data);

        data := get_matrix_element(src,y,x-hlevel);(* horizontal coefficients. *)
        put_matrix_element(dest,rd,cd-d,data);

        data := get_matrix_element(src,y-vlevel,x);(* vertical coefficients. *)
        put_matrix_element(dest,rd-d,cd,data);

        dec(cd,step);
      end;
      dec(rd,step);
    end;
    rd := src.r-1-(step-1);
    cd_max := src.c-(step-1)-1;
    d := step;
    step := step shl 1;
    inc(y_offset,vlevel);
    hlevel := hlevel shr 1;
    vlevel := vlevel shr 1;
    dec(level);
  end;

  if (hlevel<>0) and (vlevel<>0) then begin
    hlevel := src.c shr max_level;
    vlevel := src.r shr max_level;
    y_offset := src.r - vlevel;
    rd := 0;
    for y:=0 to vlevel-1 do begin
      cd := 0;
      for x:=0 to hlevel-1 do begin

        data := get_matrix_element(src,y,x);
        put_matrix_element(dest,rd,cd,data);

        inc(cd,d);
      end;
      inc(rd,d);
    end;
  end;

end;

procedure test_matrix(m: matrix);
var
  i, j: integer;
begin
  for i:=0 to m.r-1 do begin
    for j:=0 to m.c-1 do begin
      put_matrix_element(m,i,j,i);
    end;
  end;
end;

procedure clear_matrix(var m: matrix);
var
  i, j: integer;
begin
  for i:=0 to m.r-1 do begin
    for j:=0 to m.c-1 do begin
      put_matrix_element(m,i,j,0);
    end;
  end;
end;

procedure absolute_threshold_matrix(var m: matrix; threshold: element_type);
var
  row: vector;
  i: integer;
  error: boolean;
begin
  create_vector(row,m.c,error);
  for i:=0 to m.r-1 do begin
    get_row(m,i,row);
    absolute_threshold_vector(row,threshold);
    put_row(m,i,row);
  end;
  destroy_vector(row);
end;

procedure threshold_matrix(var m: matrix; threshold: element_type);
var
  row: vector;
  i: integer;
  error: boolean;
begin
  create_vector(row,m.c,error);
  for i:=0 to m.r-1 do begin
    get_row(m,i,row);
    threshold_vector(row,threshold);
    put_row(m,i,row);
  end;
  destroy_vector(row);
end;

procedure clip_matrix(var m: matrix; clip_level: element_type);
var
  row: vector;
  i: integer;
  error: boolean;
begin
  create_vector(row,m.c,error);
  for i:=0 to m.r-1 do begin
    get_row(m,i,row);
    clip_vector(row,clip_level);
    put_row(m,i,row);
  end;
  destroy_vector(row);
end;

procedure absolute_clip_matrix(var m: matrix; clip_level: element_type);
var
  row: vector;
  i: integer;
  error: boolean;
begin
  create_vector(row,m.c,error);
  for i:=0 to m.r-1 do begin
    get_row(m,i,row);
    absolute_clip_vector(row,clip_level);
    put_row(m,i,row);
  end;
  destroy_vector(row);
end;

function matrix_psnr(m, ref: matrix): real;
var
  i, j: integer;
  total, x, y, s1, s2, mse: real;
begin
  s1 := 0;
  s2 := 0;
  total := m.r;
  total := total*m.c;
  for i:=0 to m.r-1 do begin
    for j:=0 to m.c-1 do begin
      x := get_matrix_element(ref,i,j);
      y := get_matrix_element(m,i,j);
      s1 := s1 + Sqr(y-x);
      s2 := s2 + Sqr(x);
    end;
  end;
  if total<>0 then mse := s1/total
  else mse := Sqrt(255.0);
  Write('MSE=',mse:1:4,', ');
  if mse<>0 then matrix_psnr := 20.0*( Ln(255.0/Sqrt(mse))/Ln(10.0) )
  else matrix_psnr := 9999;
end;

function matrix_mean(m: matrix): real;
var
  i, j: integer;
  total, mean: real;
begin
  total := m.r;
  total := total*m.c;
  if total<>0 then begin
    mean := 0;
    for i:=0 to m.r-1 do begin
      for j:=0 to m.c-1 do begin
        mean := mean + get_matrix_element(m,i,j);
      end;
    end;
    matrix_mean := mean/total;
  end
  else matrix_mean := -1;
end;

procedure matrix_statistics(m: matrix; var stats: statistics);
var
  i, j: integer;
  mean, mean2, data: real;
begin
  stats.N := m.r;
  stats.N := stats.N*m.c;
  if stats.N<>0 then begin
    mean := 0;
    mean2 := 0;
    for i:=0 to m.r-1 do begin
      for j:=0 to m.c-1 do begin
        data := get_matrix_element(m,i,j);
        mean := mean + data;
        mean2 := mean2 + Sqr(data);
      end;
    end;
    mean := mean/stats.N;
    mean2 := mean2/stats.N;
    stats.mean := mean;
    stats.variance := mean2 - Sqr(mean);
  end
  else begin
    stats.mean := 0;
    stats.variance := -1;
  end;
end;

procedure matrix_shift(m: matrix; shift: element_type);
var
  i, j: integer;
begin
  for i:=0 to m.r-1 do begin
    for j:=0 to m.c-1 do begin
      put_matrix_element(m,i,j,get_matrix_element(m,i,j)+shift);
    end;
  end;
end;



end.
